def test_script_list(local_runner):
    res = local_runner(["script", "list"])
    assert res.code == 0, res.stderr
    assert "Found 1 script" in res.stdout, res.stderr
    assert "hello.py" in res.stdout, res.stderr

def test_script_show(local_runner):
    res = local_runner(["script", "show", "hello.py"])
    assert res.code == 0, res.stderr
    assert "print('hello world')" in res.stdout, res.stderr

def test_script_run(local_runner):
    res = local_runner(["script", "run", "hello.py"])
    assert res.code == 0, res.stderr
    assert "hello world" in res.stdout, res.stderr


from dataclasses import dataclass
import subprocess
from typing import List


@dataclass
class Result:
    stdout: str
    stderr: str
    code: int


@dataclass
class CLIRunner:
    """Dynamically create a CLI Runner for testing"""

    command: List[str]

    def __call__(self, args: List[str]) -> Result:
        proc = subprocess.Popen(
            self.command + args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = (output.decode("utf-8") for output in proc.communicate())
        return Result(out, err, proc.returncode)



import os
import shutil
import subprocess
from pathlib import Path
import sys
from typing import List, Optional
import typer

from snk.cli.dynamic_typer import DynamicTyper
from snk.cli.workflow import create_workflow
from snk.pipeline import Pipeline
from rich.console import Console
from rich.syntax import Syntax
from snakemake.deployment.conda import Conda, Env, CreateCondaEnvironmentException
from snk.cli.config.config import get_config_from_pipeline_dir


class EnvApp(DynamicTyper):
    def __init__(
        self,
        pipeline: Pipeline,
        conda_prefix_dir: Path,
        snakemake_config,
        snakefile: Path,
    ):
        self.pipeline = pipeline
        self.conda_prefix_dir = conda_prefix_dir
        self.snakemake_config = snakemake_config
        self.snakefile = snakefile
        self.configfile = get_config_from_pipeline_dir(self.pipeline.path)
        self.workflow = create_workflow(
            self.snakefile,
            config=self.snakemake_config,
            configfiles=[self.configfile] if self.configfile else None,
            use_conda=True,
            conda_prefix=self.conda_prefix_dir.resolve(),
        )
        self.register_command(self.list, help="List the environments in the pipeline.")
        self.register_command(
            self.show, help="Show the contents of an environment."
        )
        self.register_command(
            self.run, help="Run a command in one of the pipeline environments."
        )
        self.register_command(
            self.activate, help="Activate a pipeline conda environment."
        )
        self.register_command(self.prune, help="Delete all conda environments.")
        self.register_command(self.create, help="Create all conda environments.")

    def list(
            self,
            verbose: bool = typer.Option(False, "--verbose", "-v", help="Show profiles as paths."), 
        ):
        number_of_environments = len(self.pipeline.environments)
        typer.echo(
            f"Found {number_of_environments} environment{'' if number_of_environments == 1 else 's'}:"
        )
        for env in self.pipeline.environments:
            if verbose:
                typer.echo(f"- {typer.style(env, fg=typer.colors.YELLOW)}")
            else:
                typer.echo(f"- {env.stem}")

    def _get_conda_env_path(self, name: str) -> Path:
        env = [e for e in self.pipeline.environments if e.stem == name]
        if not env:
            self.error(f"Environment {name} not found!")
        return env[0]

    def _shellcmd(self, env_address: str, cmd: str) -> str:
        if sys.platform.lower().startswith("win"):
            return Conda().shellcmd_win(env_address, cmd)
        return Conda().shellcmd(env_address, cmd)

    def show(
        self, 
        name: str = typer.Argument(..., help="The name of the environment."),
        pretty: bool = typer.Option(
            False, "--pretty", "-p", help="Pretty print the environment."
        ),
    ):
        env_path = self._get_conda_env_path(name)
        env_file_text = env_path.read_text()
        if pretty:
            syntax = Syntax(env_file_text, "yaml")
            console = Console()
            console.print(syntax)
        else:
            typer.echo(env_file_text)

    def run(
        self,
        name: str = typer.Argument(..., help="The name of the environment."),
        cmd: List[str] = typer.Argument(..., help="The command to run in environment."),
    ):
        env_path = self._get_conda_env_path(name)
        env = Env(self.workflow, env_file=env_path.resolve())
        env.create()
        cmd = self._shellcmd(env.address, " ".join(cmd))
        subprocess.run(cmd, shell=True, env=os.environ.copy())

    def prune(
        self,
        force: bool = typer.Option(
            False, "--force", "-f", help="Force deletion of the environments."
        ),
    ):
        if force or input(f"Delete {self.conda_prefix_dir}? [y/N] ").lower() == "y":
            # delete self.conda_prefix_dir directory
            shutil.rmtree(self.conda_prefix_dir)
            self.log(f"Deleted {self.conda_prefix_dir}")

    def create(self):
        for env_path in self.pipeline.environments:
            env = Env(self.workflow, env_file=env_path.resolve())
            try:
                env.create()
            except CreateCondaEnvironmentException:
                self.error(
                    f"Environment {env_path.name} could not be created!", exit=False
                )

    def activate(
        self, name: str = typer.Argument(..., help="The name of the environment.")
    ):
        env_path = self._get_conda_env_path(name)
        self.log(f"Activating {name} environment... (type 'exit' to deactivate)")
        env = Env(self.workflow, env_file=env_path.resolve())
        env.create()
        user_shell = os.environ.get("SHELL", "/bin/sh")
        activate_cmd = self._shellcmd(env.address, user_shell)
        self.log(activate_cmd)
        subprocess.run(activate_cmd, shell=True, env=os.environ.copy())
        self.log(f"Exiting {name} environment...")



from pathlib import Path
from snk.cli.utils import flatten, convert_key_to_snakemake_format
from ..utils import CLIRunner


def test_config_override(local_runner: CLIRunner):
    res = local_runner(
        [
            "run",
            "--text",
            "passed from the cli to overwrite config",
            "--config",
            "tests/data/pipeline/config.yaml",
            "-f",
        ]
    )
    assert res.code == 0, res.stderr
    assert "hello_world" in res.stderr
    assert "passed from the cli to overwrite config" in res.stdout


def test_exit_on_fail(local_runner: CLIRunner):
    res = local_runner(["run", "-f", "error"])
    assert res.code == 1, res.stderr


def test_config(print_config_runner: CLIRunner):
    res = print_config_runner(["run"])
    assert res.code == 0, res.stderr
    assert "snk" in res.stdout
    res = print_config_runner(["run", "--config", "tests/data/print_value/config.yaml"])
    assert res.code == 0, res.stderr
    assert "config" in res.stdout
    res = print_config_runner(
        ["run", "--config", "tests/data/print_value/config.yaml", "--value", "cli"]
    )
    assert res.code == 0, res.stderr
    assert "cli" in res.stdout



import typer
from snk.cli.dynamic_typer import DynamicTyper
from snk.pipeline import Pipeline
from rich.console import Console
from rich.syntax import Syntax
from pathlib import Path


class ProfileApp(DynamicTyper):
    def __init__(
        self,
        pipeline: Pipeline,
    ):
        self.pipeline = pipeline
        self.register_command(self.list, help="List the profiles in the pipeline.")
        self.register_command(
            self.show, help="Show the contents of a profile."
        )

    def list(
        self,
        verbose: bool = typer.Option(False, "--verbose", "-v", help="Show profiles as paths."), 
    ):
        number_of_profiles = len(self.pipeline.profiles)
        typer.echo(
            f"Found {number_of_profiles} profile{'' if number_of_profiles == 1 else 's'}:"
        )
        for profile in self.pipeline.profiles:
            if verbose:
                typer.echo(f"- {typer.style(profile, fg=typer.colors.YELLOW)}")
            else:
                typer.echo(f"- {profile.stem}")
    
    def _get_profile_path(self, name: str) -> Path:
        profile = [e for e in self.pipeline.profiles if e.name == name or e.stem == name]
        if not profile:
            self.error(f"Profile {name} not found!")
        return profile[0]
    
    def show(
        self, 
        name: str = typer.Argument(..., help="The name of the profile."),
        pretty: bool = typer.Option(
            False, "--pretty", "-p", help="Pretty print the profile."
        ),
    ):
        profile_path = self._get_profile_path(name)
        profile_file_text = profile_path.read_text()
        if pretty:
            syntax = Syntax(profile_file_text, "yaml")
            console = Console()
            console.print(syntax)
        else:
            typer.echo(profile_file_text)



import pytest
from pathlib import Path
from snk import Nest
from .utils import CLIRunner


@pytest.fixture()
def bin_dir(tmp_path_factory: Path):
    return tmp_path_factory.mktemp("bin")


@pytest.fixture()
def snk_home(tmp_path_factory: Path):
    return tmp_path_factory.mktemp("snk")


@pytest.fixture()
def nest(snk_home, bin_dir):
    return Nest(snk_home, bin_dir)


@pytest.fixture()
def example_config():
    return Path("tests/data/config.yaml")


@pytest.fixture(scope="session")
def basic_runner(tmp_path_factory):
    nest = Nest(tmp_path_factory.mktemp("snk"), tmp_path_factory.mktemp("bin"))
    basic_pipeline = nest.install("https://github.com/Wytamma/snk-basic-pipeline.git")
    expected = nest.snk_pipelines_dir / "snk-basic-pipeline"
    assert expected.exists()
    print(basic_pipeline.executable)
    runner = CLIRunner([basic_pipeline.executable])
    return runner


@pytest.fixture(scope="session")
def local_pipeline(tmp_path_factory):
    path = Path(tmp_path_factory.mktemp("snk"))
    (path / "home").mkdir()
    (path / "bin").mkdir()
    nest = Nest(path / "home", path / "bin")
    print(nest.bin_dir)
    print(nest.snk_home)
    local_pipeline = nest.install("tests/data/pipeline", editable=True)
    expected = nest.snk_pipelines_dir / "pipeline"
    assert expected.exists()
    return local_pipeline


@pytest.fixture(scope="session")
def local_runner(tmp_path_factory):
    path = Path(tmp_path_factory.mktemp("snk"))
    (path / "home").mkdir()
    (path / "bin").mkdir()
    nest = Nest(path / "home", path / "bin")
    local_pipeline = nest.install("tests/data/pipeline", editable=True)
    runner = CLIRunner([local_pipeline.executable])
    return runner

@pytest.fixture(scope="session")
def print_config_runner(tmp_path_factory):
    path = Path(tmp_path_factory.mktemp("snk"))
    (path / "home").mkdir()
    (path / "bin").mkdir()
    nest = Nest(path / "home", path / "bin")
    local_pipeline = nest.install("tests/data/print_value", editable=True)
    runner = CLIRunner([local_pipeline.executable])
    return runner



import os
import subprocess
from pathlib import Path
import sys
from typing import List
import typer

from snk.cli.dynamic_typer import DynamicTyper
from snk.cli.workflow import create_workflow
from snk.pipeline import Pipeline
from rich.console import Console
from rich.syntax import Syntax
from snakemake.deployment.conda import Conda, Env
from snk.cli.config.config import get_config_from_pipeline_dir


class ScriptApp(DynamicTyper):
    def __init__(
        self,
        pipeline: Pipeline,
        conda_prefix_dir: Path,
        snakemake_config,
        snakefile: Path,
    ):
        self.pipeline = pipeline
        self.conda_prefix_dir = conda_prefix_dir
        self.snakemake_config = snakemake_config
        self.snakefile = snakefile
        self.configfile = get_config_from_pipeline_dir(self.pipeline.path)
        self.register_command(self.list, help="List the scripts in the pipeline.")
        self.register_command(
            self.show, help="Show the contents of a script."
        )
        self.register_command(
            self.run, 
            help=f"""
            Run a script from the pipeline.
            \n\nThe executor for the script is inferred from the suffix.
            \n\nExample: {self.pipeline.name} script run --env=python script.py
            \n\nTo pass help to the script, use `-- --help`.""",
            context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
        )

    def list(
            self, 
            verbose: bool = typer.Option(False, "--verbose", "-v", help="Show profiles as paths."), 
        ):
        number_of_scripts = len(self.pipeline.scripts)  
        typer.echo(
            f"Found {number_of_scripts} script{'' if number_of_scripts == 1 else 's'}:"
        )
        for script in self.pipeline.scripts:
            filename = typer.style(script.name, fg=typer.colors.GREEN)
            if verbose:
                filename = typer.style(script, fg=typer.colors.YELLOW)
            typer.echo(f"- {script.stem} ({filename})")

    def _get_script_path(self, name: str) -> Path:
        script = [e for e in self.pipeline.scripts if e.name == name or e.stem == name]
        if not script:
            self.error(f"Script {name} not found!")
        return script[0]

    def _get_conda_env_path(self, name: str) -> Path:
        env = [e for e in self.pipeline.environments if e.stem == name]
        if not env:
            self.error(f"Environment {name} not found!")
        return env[0]

    def _shellcmd(self, env_address: str, cmd: str) -> str:
        if sys.platform.lower().startswith("win"):
            return Conda().shellcmd_win(env_address, cmd)
        return Conda().shellcmd(env_address, cmd)

    def show(
        self, 
        name: str = typer.Argument(..., help="The name of the script."),
        pretty: bool = typer.Option(
            False, "--pretty", "-p", help="Pretty print the script."
        ),
    ):
        script_path = self._get_script_path(name)
        code = script_path.read_text()
        if pretty:
            code = Syntax(code, script_path.suffix[1:])
            console = Console()
            console.print(code)
        else:
            typer.echo(code)

    def _get_executor(self, suffix: str) -> str:
        if suffix == "py":
            return "python"
        elif suffix == "R":
            return "Rscript"
        elif suffix == "sh":
            return "bash"
        elif suffix == "pl":
            return "perl"
        elif suffix == "Rmd" or suffix == "rmd" or suffix == "Rhtml":
            return "Rscript -e 'rmarkdown::render(\"{script_path}\")'"
        elif suffix == "ipynb":
            return "jupyter nbconvert --to notebook --execute --inplace --ExecutePreprocessor.timeout=-1 {script_path}"
        elif suffix == "Rnw":
            return "Rscript -e 'knitr::knit(\"{script_path}\")'"        
        else:
            self.error(f"Unknown script suffix: {suffix}!")

    def run(
        self,
        env: str = typer.Option(None, help="The name of the environment to run script in."),
        name: str = typer.Argument(..., help="The name of the script."),
        args: List[str] = typer.Argument(None, help="Arguments to pass to the script."),
    ):
        script_path = self._get_script_path(name)
        executor = self._get_executor(script_path.suffix[1:])
        cmd = [executor, str(script_path)] + args
        if env:
            env_path = self._get_conda_env_path(env)
            workflow = create_workflow(
                self.snakefile,
                config=self.snakemake_config,
                configfiles=[self.configfile] if self.configfile else None,
                use_conda=True,
                conda_prefix=self.conda_prefix_dir.resolve(),
            )
            env = Env(workflow, env_file=env_path.resolve())
            env.create()
            cmd = self._shellcmd(env.address, " ".join(cmd))
        else:
            cmd = " ".join(cmd)
        subprocess.run(cmd, shell=True, env=os.environ.copy())





import pytest
from snk.cli.config import SnkConfig
from snk.cli.options import Option
from snk.cli.options.utils import (
    create_option_from_annotation,
    build_dynamic_cli_options,
)


@pytest.fixture
def default_annotation_values():
    return {
        "test:name": "Test",
        "test:default": "default_value",
        "test:type": "str",
        "test:help": "Test help",
        "test:required": True,
    }


@pytest.fixture
def default_default_values():
    return {"test": "default_value"}


def test_create_option_from_annotation(
    default_annotation_values, default_default_values
):
    option = create_option_from_annotation(
        "test", default_annotation_values, default_default_values
    )

    assert isinstance(option, Option)
    assert option.name == "Test"
    assert option.default == "default_value"
    assert option.updated == False
    assert option.help == "Test help"
    assert option.type == str
    assert option.required == True


# @pytest.mark.parametrize("annotation_values, default_values, expected_type", [
#     ({"test:type": "str"}, {}, str),
# ])
# def test_nested_option():
#     pass


@pytest.fixture
def default_snakemake_config():
    return {
        "param1": "value1",
        "param2": "value2",
    }


@pytest.fixture
def default_snk_config():
    return SnkConfig(
        {
            "annotations": {
                "param1:name": "Parameter 1",
                "param2:name": "Parameter 2",
            }
        }
    )


def test_build_dynamic_cli_options(default_snakemake_config, default_snk_config):
    options = build_dynamic_cli_options(default_snakemake_config, default_snk_config)

    assert len(options) == 2
    assert all(isinstance(option, Option) for option in options)
    assert set([option.name for option in options]) == {"param1", "param2"}



import pytest
from typer.testing import CliRunner
from snk.cli.dynamic_typer import DynamicTyper, Option
from snk.cli.options import Option
from typing import Callable
from inspect import signature, Parameter
import typer


@pytest.fixture
def dynamic_typer():
    return DynamicTyper()

@pytest.fixture
def dynamic_app():
    class App(DynamicTyper):
        pass
    return App()

@pytest.fixture
def cli_runner():
    return CliRunner()


def test_dynamic_typer_init(dynamic_app):
    assert isinstance(dynamic_app.app, typer.Typer)


def test_register_default_command(dynamic_app: DynamicTyper, cli_runner):
    
    class App(DynamicTyper):
        def __init__(self):
            self.register_default_command(self.hello)
        def hello(self):
            self.echo("Hello World")

    result = cli_runner.invoke(App().app)
    assert result.exit_code == 0, result.stdout
    assert result.stdout.strip() == "Hello World"


def test_register_command(dynamic_app: DynamicTyper, cli_runner):
    def command(name: str):
        print(f"Hello {name}")

    dynamic_app.register_command(command)
    result = cli_runner.invoke(dynamic_app.app, ["John"])
    assert result.exit_code == 0
    assert result.stdout.strip() == "Hello John"


def test_register_callback(dynamic_app: DynamicTyper, cli_runner):
    
    def callback(name: str):
        print(f"Hello {name}")

    dynamic_app.register_callback(callback, invoke_without_command=True)
    result = cli_runner.invoke(dynamic_app.app, ["John"])
    assert result.exit_code == 0, result.stdout
    assert result.stdout.strip() == "Hello John"


def test_register_group(dynamic_app: DynamicTyper, cli_runner):
    
    class SubApp(DynamicTyper):
        def __init__(self):
            self.register_command(self.hello)
        def hello(self):
            self.echo("Hello World")

    dynamic_app.register_group(SubApp(), name="subapp", help="A subapp")

    result = cli_runner.invoke(dynamic_app.app, ["subapp", "hello"])
    assert result.exit_code == 0, result.stdout
    assert result.stdout.strip() == "Hello World"


def test_create_cli_parameter(dynamic_typer):
    option = Option(
        name="foo",
        type=int,
        required=True,
        default=0,
        help="A number",
        short="f",
        updated=True,
        original_key="foo",
        flag="--foo",
        short_flag="-f",
    )
    param = dynamic_typer._create_cli_parameter(option)
    assert isinstance(param, Parameter)
    assert param.name == "foo"
    assert param.annotation == int
    assert isinstance(param.default, typer.models.OptionInfo)
    assert param.kind == Parameter.POSITIONAL_OR_KEYWORD



def test_add_dynamic_options(dynamic_typer):
    def func(name: str):
        return f"Hello {name}"

    options = [
        Option(
            name="foo",
            type=int,
            required=True,
            default=0,
            help="A number",
            short="f",
            updated=True,
            original_key="foo",
            flag="--foo",
            short_flag="-f",
        )
    ]
    func_wrapper = dynamic_typer.add_dynamic_options(func, options)
    assert callable(func_wrapper)
    assert "foo" in signature(func_wrapper).parameters


def test_error(dynamic_typer, cli_runner):
    with pytest.raises(typer.Exit):
        dynamic_typer.error("Error message")


def test_log(dynamic_typer, cli_runner, capsys):
    dynamic_typer.log("Log message")
    captured = capsys.readouterr()
    assert "Log message" in captured.out



import inspect
import platform

import typer
from pathlib import Path
from typing import Optional
import os


from snakemake import SNAKEFILE_CHOICES
from art import text2art

from snk.cli.dynamic_typer import DynamicTyper
from snk.cli.subcommands import EnvApp, ConfigApp, RunApp, ScriptApp, ProfileApp

from .config.config import (
    SnkConfig,
    load_pipeline_snakemake_config,
)
from .options.utils import build_dynamic_cli_options
from snk.pipeline import Pipeline


class CLI(DynamicTyper):
    """
    Constructor for the dynamic Snk CLI class.
    Args:
      pipeline_dir_path (Path): Path to the pipeline directory.
    Side Effects:
      Initializes the CLI class.
    Examples:
      >>> CLI(Path('/path/to/pipeline'))
    """

    def __init__(self, pipeline_dir_path: Path = None) -> None:
        if not pipeline_dir_path:
            # get the calling frame (the frame of the function that called this function)
            calling_frame = inspect.currentframe().f_back
            # get the file path from the calling frame
            pipeline_dir_path = Path(calling_frame.f_globals["__file__"])
        if pipeline_dir_path.is_file():
            pipeline_dir_path = pipeline_dir_path.parent
        self.pipeline = Pipeline(path=pipeline_dir_path)
        self.snakemake_config = load_pipeline_snakemake_config(pipeline_dir_path)
        self.snk_config = SnkConfig.from_pipeline_dir(
            pipeline_dir_path, create_if_not_exists=True
        )
        if self.snk_config.version:
            self.version = self.snk_config.version
        else: 
            if self.pipeline.tag:
               self.version = self.pipeline.tag
            else:
                self.version = self.pipeline.commit
        self.options = build_dynamic_cli_options(self.snakemake_config, self.snk_config)
        self.snakefile = self._find_snakefile()
        self.conda_prefix_dir = pipeline_dir_path / ".conda"
        if " " in str(pipeline_dir_path):
            # cannot have spaces!
            self.singularity_prefix_dir = None
        else:
            self.singularity_prefix_dir = pipeline_dir_path / ".singularity"
        self.name = self.pipeline.name
        self.verbose = False
        if (
            platform.system() == "Darwin"
            and platform.processor() == "arm"
            and not os.environ.get("CONDA_SUBDIR")
        ):
            os.environ["CONDA_SUBDIR"] = "osx-64"

        # dynamically create the logo
        self.logo = self._create_logo(
            tagline=self.snk_config.tagline, font=self.snk_config.font
        )
        callback = self._create_callback()
        callback.__doc__ = self.logo

        # registration
        self.register_callback(
            callback,
            invoke_without_command=True,
            context_settings={"help_option_names": ["-h", "--help"]},
        )
        self.register_command(self.info, help="Show information about the pipeline.")

        run_app = RunApp(
            conda_prefix_dir=self.conda_prefix_dir,
            snk_config=self.snk_config,
            singularity_prefix_dir=self.singularity_prefix_dir,
            snakefile=self.snakefile,
            pipeline=self.pipeline,
            verbose=self.verbose,
            logo=self.logo,
            dynamic_run_options=self.options,
        )
        # Subcommands
        self.register_command(
            run_app,
            name="run",
        )
        self.register_command(
            ConfigApp(
                pipeline=self.pipeline,
                options=self.options,
            ),
            name="config",
        )
        if self.pipeline.environments:
            self.register_group(
                EnvApp(
                    pipeline=self.pipeline,
                    conda_prefix_dir=self.conda_prefix_dir,
                    snakemake_config=self.snakemake_config,
                    snakefile=self.snakefile,
                ),
                name="env",
                help="Access the pipeline conda environments.",
            )
        if self.pipeline.scripts:
            self.register_group(
                ScriptApp(
                    pipeline=self.pipeline,
                    conda_prefix_dir=self.conda_prefix_dir,
                    snakemake_config=self.snakemake_config,
                    snakefile=self.snakefile,
                ),
                name="script",
                help="Access the pipeline scripts.",
            )
        if self.pipeline.profiles:
            self.register_group(
                ProfileApp(
                    pipeline=self.pipeline,
                ),
                name="profile",
                help="Access the pipeline profiles.",
            )

    def _print_pipline_version(self, ctx: typer.Context, value: bool):
        if value:
            typer.echo(self.version)
            raise typer.Exit()

    def _print_pipline_path(self, ctx: typer.Context, value: bool):
        if value:
            typer.echo(self.pipeline.path)
            raise typer.Exit()

    def _create_callback(self):
        def callback(
            ctx: typer.Context,
            version: Optional[bool] = typer.Option(
                None,
                "-v",
                "--version",
                help="Show the pipeline version and exit.",
                is_eager=True,
                callback=self._print_pipline_version,
                show_default=False,
            ),
            path: Optional[bool] = typer.Option(
                None,
                "-p",
                "--path",
                help="Show the pipeline path and exit.",
                is_eager=True,
                callback=self._print_pipline_path,
                show_default=False,
            ),
        ):
            if ctx.invoked_subcommand is None:
                typer.echo(f"{ctx.get_help()}")

        return callback

    def _create_logo(
        self, tagline="A Snakemake pipeline CLI generated with snk", font="small"
    ):
        """
        Create a logo for the CLI.
        Args:
          font (str): The font to use for the logo.
        Returns:
          str: The logo.
        Examples:
          >>> CLI._create_logo()
        """
        if self.snk_config.art:
            art = self.snk_config.art
        else:
            logo = self.snk_config.logo if self.snk_config.logo else self.name
            art = text2art(logo, font=font)
        doc = f"""\b{art}\b{tagline}"""
        return doc

    def _find_snakefile(self):
        """
        Search possible snakefile locations.
        Returns:
          Path: The path to the snakefile.
        Examples:
          >>> CLI._find_snakefile()
        """
        for path in SNAKEFILE_CHOICES:
            if (self.pipeline.path / path).exists():
                return self.pipeline.path / path
        raise FileNotFoundError("Snakefile not found!")

    def info(self):
        """
        Display information about current pipeline install.
        Returns:
          str: A JSON string containing information about the current pipeline install.
        Examples:
          >>> CLI.info()
        """
        import json

        info_dict = {}
        info_dict["name"] = self.pipeline.path.name
        info_dict["version"] = self.version
        info_dict["pipeline_dir_path"] = str(self.pipeline.path)
        typer.echo(json.dumps(info_dict, indent=2))




from pathlib import Path
from typer.testing import CliRunner

from snk.main import app
from snk.pipeline import Pipeline

runner = CliRunner()


def test_snk_help():
    result = runner.invoke(app, ["-h"])
    assert result.exit_code == 0
    assert "snk" in result.stdout
    assert "Usage" in result.stdout


def test_snk_install(snk_home: Path, bin_dir: Path):
    result = runner.invoke(
        app, ["--home", snk_home, "--bin", bin_dir, "install", "tests/data/pipeline"]
    )
    assert result.exit_code == 0
    assert "Successfully installed" in result.stdout
    assert (snk_home / "pipelines" / "pipeline").exists()
    assert (bin_dir / "pipeline").is_symlink()


def test_snk_install_no_conda(snk_home: Path, bin_dir: Path):
    result = runner.invoke(
        app, ["--home", snk_home, "--bin", bin_dir, "install", "tests/data/pipeline", "--no-conda"]

    )
    assert result.exit_code == 0
    assert "Successfully installed" in result.stdout
    snk_config = snk_home / "pipelines" / "pipeline"/ "snk.yaml"
    assert snk_config.exists()
    # read the snk.yaml file as dict
    import yaml
    with open(snk_config) as f:
        snk_config_dict = yaml.safe_load(f)
    assert snk_config_dict["conda"] is False


def test_snk_list(local_pipeline: Pipeline):
    snk_home = local_pipeline.path.parent.parent
    bin_dir = local_pipeline.path.parent.parent.parent / "bin"
    result = runner.invoke(app, ["--home", snk_home, "--bin", bin_dir, "list"])
    assert result.exit_code == 0
    assert "Found 1 pipelines" in result.stdout


def test_snk_uninstall(local_pipeline: Pipeline):
    snk_home = local_pipeline.path.parent.parent
    bin_dir = local_pipeline.path.parent.parent.parent / "bin"
    result = runner.invoke(
        app, ["--home", snk_home, "--bin", bin_dir, "uninstall", "pipeline"]
    )
    assert "(Y/n)" in result.stdout
    result = runner.invoke(
        app, ["--home", snk_home, "--bin", bin_dir, "uninstall", "pipeline", "--force"]
    )
    assert result.exit_code == 0
    assert "Successfully uninstalled" in result.stdout



from typing import List
import typer

from snk.cli.dynamic_typer import DynamicTyper
from snk.cli.options.option import Option
from snk.pipeline import Pipeline


class ConfigApp(DynamicTyper):
    def __init__(self, pipeline: Pipeline, options: List[Option]):
        """
        Initializes the ConfigApp class.
        Args:
            pipeline (Pipeline): The pipeline to configure.
        """
        self.options = options
        self.pipeline = pipeline
        self.register_command(self.config, help="Show the pipeline configuration.")

    def config(
        self, ctx: typer.Context, pretty: bool = typer.Option(False, "--pretty", "-p")
    ):
        """
        Prints the configuration for the pipeline.
        Args:
            pretty (bool, optional): Whether to print the configuration in a pretty format. Defaults to False.
        Returns:
            None
        Examples:
            >>> ConfigApp.show(pretty=True)
            # Pretty printed configuration
        """
        import yaml
        from collections import defaultdict
        from rich.console import Console
        from rich.syntax import Syntax
        from snk.cli.utils import convert_key_to_snakemake_format

        def deep_update(source, overrides):
            for key, value in overrides.items():
                if isinstance(value, dict):
                    if not isinstance(source.get(key), dict):
                        # If the existing value is not a dictionary, replace it with one
                        source[key] = {}
                    # Now we are sure that source[key] is a dictionary, so we can update it
                    deep_update(source[key], value)
                else:
                    source[key] = value
            return source

        collapsed_data = defaultdict(dict)
        config_dict = [
            convert_key_to_snakemake_format(option.original_key, option.default)
            for option in self.options
        ]
        for d in config_dict:
            deep_update(collapsed_data, d)
        yaml_str = yaml.dump(dict(collapsed_data))
        if pretty:
            syntax = Syntax(yaml_str, "yaml")
            console = Console()
            console.print(syntax)
        else:
            typer.echo(yaml_str)



from pathlib import Path
from typing import List, Optional
import snakemake
from dataclasses import dataclass, field
from snk.cli.config.utils import get_version_from_config
from snk.errors import InvalidSnkConfigError, MissingSnkConfigError
import yaml

@dataclass
class SnkConfig:
    """
    A dataclass for storing Snakemake pipeline configuration.
    """

    art: str = None
    logo: str = None
    tagline: str = "A Snakemake pipeline CLI generated with Snk"
    font: str = "small"
    version: Optional[str] = None
    conda: bool = True
    resources: List[Path] = field(default_factory=list)
    cli: dict = field(default_factory=dict)
    symlink_resources: bool = False
    _snk_config_path: Path = None

    @classmethod
    def from_path(cls, snk_config_path: Path):
        """
        Load and validate Snk config from file.
        Args:
          snk_config_path (Path): Path to the SNK config file.
        Returns:
          SnkConfig: A SnkConfig object.
        Raises:
          FileNotFoundError: If the SNK config file is not found.
        Examples:
          >>> SnkConfig.from_path(Path("snk.yaml"))
          SnkConfig(art=None, logo=None, tagline='A Snakemake pipeline CLI generated with Snk', font='small', resources=[], annotations={}, symlink_resources=False, _snk_config_path=PosixPath('snk.yaml'))
        """
        if not snk_config_path.exists():
            raise MissingSnkConfigError(
                f"Could not find SNK config file: {snk_config_path}"
            ) from FileNotFoundError
        # raise error if file is empty
        if snk_config_path.stat().st_size == 0:
            raise InvalidSnkConfigError(f"SNK config file is empty: {snk_config_path}") from ValueError

        snk_config_dict = snakemake.load_configfile(snk_config_path)
        snk_config_dict["version"] = get_version_from_config(snk_config_path, snk_config_dict)
        if "annotations" in snk_config_dict:
            # TODO: remove annotations in the future
            snk_config_dict["cli"] = snk_config_dict["annotations"]
            del snk_config_dict["annotations"]
        if "conda_required" in snk_config_dict:
            # TODO: remove conda_required in the future
            snk_config_dict["conda"] = snk_config_dict["conda_required"]
            del snk_config_dict["conda_required"]
        snk_config = cls(**snk_config_dict)
        snk_config.resources = [
            snk_config_path.parent / resource for resource in snk_config.resources
        ]
        snk_config.validate_resources(snk_config.resources)
        snk_config._snk_config_path = snk_config_path
        return snk_config
  
    @classmethod
    def from_pipeline_dir(
        cls, pipeline_dir_path: Path, create_if_not_exists: bool = False
    ):
        """
        Load and validate SNK config from pipeline directory.
        Args:
          pipeline_dir_path (Path): Path to the pipeline directory.
          create_if_not_exists (bool): Whether to create a SNK config file if one does not exist.
        Returns:
          SnkConfig: A SnkConfig object.
        Raises:
          FileNotFoundError: If the SNK config file is not found.
        Examples:
          >>> SnkConfig.from_pipeline_dir(Path("pipeline"))
          SnkConfig(art=None, logo=None, tagline='A Snakemake pipeline CLI generated with Snk', font='small', resources=[], annotations={}, symlink_resources=False, _snk_config_path=PosixPath('pipeline/snk.yaml'))
        """
        if (pipeline_dir_path / "snk.yaml").exists():
            return cls.from_path(pipeline_dir_path / "snk.yaml")
        elif (pipeline_dir_path / ".snk").exists():
            import warnings

            warnings.warn(
                "Use of .snk will be deprecated in the future. Please use snk.yaml instead.",
                DeprecationWarning,
            )
            return cls.from_path(pipeline_dir_path / ".snk")
        elif create_if_not_exists:
            snk_config = cls(_snk_config_path=pipeline_dir_path / "snk.yaml")
            return snk_config
        else:
            raise FileNotFoundError(
                f"Could not find SNK config file in pipeline directory: {pipeline_dir_path}"
            )

    def validate_resources(self, resources):
        """
        Validate resources.
        Args:
          resources (List[Path]): List of resources to validate.
        Raises:
          FileNotFoundError: If a resource is not found.
        Notes:
          This function does not modify the resources list.
        Examples:
          >>> SnkConfig.validate_resources([Path("resource1.txt"), Path("resource2.txt")])
        """
        for resource in resources:
            if not resource.exists():
                raise FileNotFoundError(f"Could not find resource: {resource}")

    def add_resources(self, resources: List[Path], pipeline_dir_path: Path = None):
        """
        Add resources to the SNK config.
        Args:
          resources (List[Path]): List of resources to add.
          pipeline_dir_path (Path): Path to the pipeline directory.
        Returns:
          None
        Side Effects:
          Adds the resources to the SNK config.
        Examples:
          >>> snk_config = SnkConfig()
          >>> snk_config.add_resources([Path("resource1.txt"), Path("resource2.txt")], Path("pipeline"))
        """
        processed = []
        for resource in resources:
            if pipeline_dir_path and not resource.is_absolute():
                resource = pipeline_dir_path / resource
            processed.append(resource)
        self.validate_resources(processed)
        self.resources.extend(processed)

    def to_yaml(self, path: Path) -> None:
        """
        Write SNK config to YAML file.
        Args:
          path (Path): Path to write the YAML file to.
        Returns:
          None
        Side Effects:
          Writes the SNK config to the specified path.
        Examples:
          >>> snk_config = SnkConfig()
          >>> snk_config.to_yaml(Path("snk.yaml"))
        """
        config_dict = {k: v for k, v in vars(self).items() if not k.startswith("_")}
        with open(path, "w") as f:
            yaml.dump(config_dict, f)

    def save(self) -> None:
        """
        Save SNK config.
        Args:
          path (Path): Path to write the YAML file to.
        Returns:
          None
        Side Effects:
          Writes the SNK config to the path specified by _snk_config_path.
        Examples:
          >>> snk_config = SnkConfig()
          >>> snk_config.save()
        """
        self.to_yaml(self._snk_config_path)


def get_config_from_pipeline_dir(pipeline_dir_path: Path):
    """
    Get the config file from a pipeline directory.
    Args:
      pipeline_dir_path (Path): Path to the pipeline directory.
    Returns:
      Path: Path to the config file, or None if not found.
    Examples:
      >>> get_config_from_pipeline_dir(Path("pipeline"))
      PosixPath('pipeline/config.yaml')
    """
    for path in [
        Path("config") / "config.yaml",
        Path("config") / "config.yml",
        "config.yaml",
        "config.yml",
    ]:
        if (pipeline_dir_path / path).exists():
            return pipeline_dir_path / path
    return None


def load_pipeline_snakemake_config(pipeline_dir_path: Path):
    """
    Load the Snakemake config from a pipeline directory.
    Args:
      pipeline_dir_path (Path): Path to the pipeline directory.
    Returns:
      dict: The Snakemake config.
    Examples:
      >>> load_pipeline_snakemake_config(Path("pipeline"))
      {'inputs': {'data': 'data.txt'}, 'outputs': {'results': 'results.txt'}}
    """
    pipeline_config_path = get_config_from_pipeline_dir(pipeline_dir_path)
    if not pipeline_config_path or not pipeline_config_path.exists():
        return {}
    return snakemake.load_configfile(pipeline_config_path)



from pathlib import Path
from snakemake import load_configfile

def get_version_from_config(config_path: Path, config_dict: dict = None) -> str:
    """
    Get the version from config. If dict not provided, load from file.
    If the version is a path to a __about__ file, load the version from the file.
    Path must be relative to the config file.
    Args:
      config_path (Path): Path to the config file.
      config_dict (dict): Config dict.
    Returns:
      str: Version.
    Examples:
      >>> get_version_from_config_dict({"version": "0.1.0"})
      '0.1.0'
    """
    if not config_dict:
        config_dict = load_configfile(config_path)
    
    if "version" not in config_dict:
        return None 
    if config_dict["version"] is None:
        return None
    version = str(config_dict["version"])
    if "__about__.py" in version:
        # load version from __about__.py
        about_path = config_path.parent / version
        if not about_path.exists():
            raise FileNotFoundError(
                f"Could not find version file: {about_path}"
            )
        about = {}
        exec(about_path.read_text(), about)
        try:
            version = about["__version__"]
        except KeyError as e:
            raise KeyError(
                f"Could not find __version__ in file: {about_path}"
            ) from e
    return version


from pathlib import Path
from snk.cli.utils import flatten, convert_key_to_snakemake_format
import snakemake
import pytest
from ..utils import CLIRunner


def test_flatten(example_config: Path):
    config = snakemake.load_configfile(example_config)
    flat_config = flatten(config)
    assert flat_config["diffexp:contrasts:A-vs-B"] == ["A", "B"]


@pytest.mark.parametrize(
    "key, value, expected",
    [
        ("a:b:c", 42, {"a": {"b": {"c": 42}}}),
        ("x:y:z", "hello", {"x": {"y": {"z": "hello"}}}),
        ("d:e", [1, 2, 3], {"d": {"e": [1, 2, 3]}}),
        ("h", "world", {"h": "world"}),
    ],
)
def test_convert_key_to_snakemake_format(key, value, expected):
    assert convert_key_to_snakemake_format(key, value) == expected


def test_help(local_runner: CLIRunner):
    res = local_runner(["--help"])
    assert res.code == 0, res.stderr
    assert "Usage:" in res.stdout

def test_version(local_runner: CLIRunner):
    res = local_runner(["--version"])
    assert res.code == 0, res.stderr
    assert "0.13.0" in res.stdout

def test_path(local_runner: CLIRunner):
    res = local_runner(["--path"])
    assert res.code == 0, res.stderr
    assert "/pipeline" in res.stdout

def test_profile(local_runner: CLIRunner):
    res = local_runner(["profile", "-h"])
    assert res.code == 0, res.stderr
    assert "list" in res.stdout, res.stderr
    assert "show" in res.stdout, res.stderr

def test_info(local_runner: CLIRunner):
    res = local_runner(["info"])
    assert res.code == 0, res.stderr
    assert "pipeline" in res.stdout, res.stderr
    assert "version" in res.stdout
    assert "pipeline_dir_path" in res.stdout

def test_config(local_runner: CLIRunner):
    res = local_runner(["config"])
    assert res.code == 0, res.stderr
    assert "output: hello.txt" in res.stdout

def test_run_cli(local_runner: CLIRunner):
    res = local_runner(["run", "-h"])
    assert res.code == 0, res.stderr
    assert "pipeline" in res.stdout, res.stderr
    assert "times" in res.stdout
    assert "text" in res.stdout


@pytest.mark.parametrize("filetype", ["pdf", "svg", "png"])
def test_dag(local_runner: CLIRunner, tmp_path, filetype):
    res = local_runner(["run", "--dag", f"{tmp_path}/dag.{filetype}"])
    assert res.code == 0, res.stderr
    assert Path(f"{tmp_path}/dag.{filetype}").exists()



from typing import List, Optional
from pathlib import Path
import typer
from contextlib import contextmanager

from snk.cli.dynamic_typer import DynamicTyper
from snk.cli.options.option import Option
from snk.pipeline import Pipeline
from snk.cli.utils import (
    parse_config_args,
    dag_filetype_callback,
)

from snk.cli.config.config import (
    SnkConfig,
    get_config_from_pipeline_dir,
)


class RunApp(DynamicTyper):
    def __init__(
        self,
        conda_prefix_dir: Path,
        snk_config: SnkConfig,
        singularity_prefix_dir: Path,
        snakefile: Path,
        pipeline: Pipeline,
        logo: str,
        verbose: bool,
        dynamic_run_options: List[Option],
    ):
        self.conda_prefix_dir = conda_prefix_dir
        self.singularity_prefix_dir = singularity_prefix_dir
        self.snk_config = snk_config
        self.snakefile = snakefile
        self.pipeline = pipeline
        self.verbose = verbose
        self.logo = logo
        self.options = dynamic_run_options

        self.register_command(
            self.run,
            dynamic_options=self.options,
            help="Run the pipeline.\n\nAll unrecognized arguments are passed onto Snakemake.",
            context_settings={
                "allow_extra_args": True,
                "ignore_unknown_options": True,
                "help_option_names": ["-h", "--help"],
            },
        )

    def _print_snakemake_help(value: bool):
        """
        Print the snakemake help and exit.
        Args:
          value (bool): If True, print the snakemake help and exit.
        Side Effects:
          Prints the snakemake help and exits.
        Examples:
          >>> CLI._print_snakemake_help(True)
        """
        if value:
            import snakemake

            snakemake.main("-h")

    def run(
        self,
        ctx: typer.Context,
        target: str = typer.Argument(
            None, help="File(s) to generate. If None will run the pipeline 'all' rule."
        ),
        configfile: Path = typer.Option(
            None,
            "--config",
            help="Path to snakemake config file. Overrides existing config and defaults.",
            exists=True,
            dir_okay=False,
        ),
        resource: List[Path] = typer.Option(
            [],
            "--resource",
            "-r",
            help="Additional resources to copy to workdir at run time (relative to pipeline directory).",
        ),
        profile: Optional[str] = typer.Option(
            None,
            "--profile",
            "-p",
            help="Name of profile to use for configuring Snakemake.",
        ),
        force: bool = typer.Option(
            False,
            "--force",
            "-f",
            help="Force the execution of pipeline regardless of already created output.",
        ),
        dry: bool = typer.Option(
            False,
            "--dry",
            "-n",
            help="Do not execute anything, and display what would be done. If you have a very large workflow, use --dry --quiet to just print a summary of the DAG of jobs",
        ),
        lock: bool = typer.Option(
            False, "--lock", "-l", help="Lock the working directory."
        ),
        keep_resources: bool = typer.Option(
            False,
            "--keep-resources",
            "-R",
            help="Keep resources after pipeline completes.",
        ),
        keep_snakemake: bool = typer.Option(
            False,
            "--keep-snakemake",
            "-S",
            help="Keep .snakemake folder after pipeline completes.",
        ),
        dag: Optional[Path] = typer.Option(
            None,
            "--dag",
            "-d",
            help="Save directed acyclic graph to file. Must end in .pdf, .png or .svg",
            callback=dag_filetype_callback,
        ),
        cores: int = typer.Option(
            None,
            "--cores",
            "-c",
            help="Set the number of cores to use. If None will use all cores.",
        ),
        verbose: Optional[bool] = typer.Option(
            False,
            "--verbose",
            "-v",
            help="Run pipeline in verbose mode.",
        ),
        help_snakemake: Optional[bool] = typer.Option(
            False,
            "--help-snakemake",
            "-hs",
            help="Print the snakemake help and exit.",
            is_eager=True,
            callback=_print_snakemake_help,
            show_default=False,
        ),
    ):
        """
        Run the pipeline.
        Args:
          target (str): File to generate. If None will run the pipeline 'all' rule.
          configfile (Path): Path to snakemake config file. Overrides existing config and defaults.
          resource (List[Path]): Additional resources to copy to workdir at run time.
          keep_resources (bool): Keep resources.
          cleanup_snakemake (bool): Keep .snakemake folder.
          cores (int): Set the number of cores to use. If None will use all cores.
          verbose (bool): Run pipeline in verbose mode.
          help_snakemake (bool): Print the snakemake help and exit.
        Side Effects:
          Runs the pipeline.
        Examples:
          >>> CLI.run(target='my_target', configfile=Path('/path/to/config.yaml'), resource=[Path('/path/to/resource')], verbose=True)
        """
        import snakemake
        import shutil
        import sys

        self.verbose = verbose
        args = []
        if not cores:
            cores = "all"
        args.extend(
            [
                "--rerun-incomplete",
                f"--cores={cores}",
            ]
        )
        if target:
            ctx.args.append(target)
        if self.singularity_prefix_dir and "--use-singularity" in ctx.args:
            # only set prefix if --use-singularity is explicitly called
            args.append(f"--singularity-prefix={self.singularity_prefix_dir}")
            if verbose:
                self.log(f"Using singularity prefix: {self.singularity_prefix_dir}")
        if not self.snakefile.exists():
            raise ValueError("Could not find Snakefile")  # this should occur at install
        else:
            args.append(f"--snakefile={self.snakefile}")

        if not configfile:
            configfile = get_config_from_pipeline_dir(self.pipeline.path)
        if configfile:
            args.append(f"--configfile={configfile}")

        if profile:
            found_profile = [p for p in self.pipeline.profiles if profile == p.name]
            if found_profile:
                profile = found_profile[0]
            args.append(f"--profile={profile}")

        # Set up conda frontend
        conda_found = check_command_available("conda")
        if not conda_found and verbose:
            typer.secho(
                "Conda not found! Install conda to use environments.\n",
                fg=typer.colors.MAGENTA,
            )

        if conda_found and self.snk_config.conda:
            args.extend([
                    "--use-conda",
                    f"--conda-prefix={self.conda_prefix_dir}",
                ]
            )
            if not check_command_available("mamba"):
                if verbose:
                    typer.secho(
                        "Could not find mamba, using conda instead...",
                        fg=typer.colors.MAGENTA,
                    )
                args.append("--conda-frontend=conda")
            else:
                args.append("--conda-frontend=mamba")

        if verbose:
            args.insert(0, "--verbose")

        if force:
            args.append("--forceall")
        
        if dry:
            args.append("--dryrun")

        if not lock:
            args.append("--nolock")

        targets_and_or_snakemake, config_dict_list = parse_config_args(
            ctx.args, options=self.options
        )
        targets_and_or_snakemake = [
            t.replace("--snake-", "-") for t in targets_and_or_snakemake
        ]
        args.extend(targets_and_or_snakemake)
        configs = []
        for config_dict in config_dict_list:
            for key, value in config_dict.items():
                configs.append(f"{key}={value}")

        if configs:
            args.extend(["--config", *configs])
        if verbose:
            typer.secho(f"snakemake {' '.join(args)}\n", fg=typer.colors.MAGENTA)
        if not keep_snakemake and Path(".snakemake").exists():
            keep_snakemake = True
        try:
            self.snk_config.add_resources(resource, self.pipeline.path)
        except FileNotFoundError as e:
            self.error(str(e))
        with self._copy_resources(
            self.snk_config.resources,
            cleanup=not keep_resources,
            symlink_resources=self.snk_config.symlink_resources,
        ):
            if dag:
                return self._save_dag(snakemake_args=args, filename=dag)
            try:
                snakemake.parse_config = parse_config_monkeypatch
                snakemake.main(args)
            except SystemExit as e:
                status = int(str(e))
                if status:
                    sys.exit(status)
        if not keep_snakemake and Path(".snakemake").exists():
            if verbose:
                typer.secho("Deleting '.snakemake' folder...", fg=typer.colors.MAGENTA)
            shutil.rmtree(".snakemake")

    def _save_dag(self, snakemake_args: List[str], filename: Path):
        from contextlib import redirect_stdout
        import snakemake
        import subprocess
        import io

        snakemake_args.append("--dag")

        fileType = filename.suffix.lstrip(".")

        # Create a file-like object to redirect the stdout
        snakemake_output = io.StringIO()
        # Use redirect_stdout to redirect stdout to the file-like object
        with redirect_stdout(snakemake_output):
            # Capture the output of snakemake.main(args) using a try-except block
            try:
                snakemake.parse_config = parse_config_monkeypatch
                snakemake.main(snakemake_args)
            except SystemExit:  # Catch SystemExit exception to prevent termination
                pass
        try:
            snakemake_output = snakemake_output.getvalue()
            if "snakemake_dag" not in snakemake_output:
                self.error("Could not generate dag!", exit=True)
            # discard everything before digraph snakemake_dag
            filtered_lines = "digraph snakemake_dag" + snakemake_output.split("snakemake_dag")[1]
            echo_process = subprocess.Popen(
                ["echo", filtered_lines], stdout=subprocess.PIPE
            )
            dot_process = subprocess.Popen(
                ["dot", f"-T{fileType}"],
                stdin=echo_process.stdout,
                stdout=subprocess.PIPE,
            )
            with open(filename, "w") as output_file:
                if self.verbose:
                    typer.secho(f"Saving dag to {filename}", fg=typer.colors.MAGENTA)
                subprocess.run(["cat"], stdin=dot_process.stdout, stdout=output_file)
        except (subprocess.CalledProcessError, FileNotFoundError):
            typer.echo("dot command not found!", fg=typer.colors.RED, err=True)
            raise typer.Exit(1)

    @contextmanager
    def _copy_resources(
        self, resources: List[Path], cleanup: bool, symlink_resources: bool = False
    ):
        """
        Copy resources to the current working directory.
        Args:
          resources (List[Path]): A list of paths to the resources to copy.
          cleanup (bool): If True, the resources will be removed after the function exits.
        Side Effects:
          Copies the resources to the current working directory.
        Returns:
          Generator: A generator object.
        Examples:
          >>> with CLI.copy_resources(resources, cleanup=True):
          ...     # do something
        """
        import os
        import shutil

        copied_resources = []

        def copy_resource(src: Path, dst: Path, symlink: bool = False):
            if self.verbose:
                typer.secho(
                    f"  - Copying resource '{src}' to '{dst}'",
                    fg=typer.colors.MAGENTA,
                )
            target_is_directory = src.is_dir()
            if symlink:
                os.symlink(src, dst, target_is_directory=target_is_directory)
            elif target_is_directory:
                shutil.copytree(src, dst)
            else:
                shutil.copy(src, dst)

        def remove_resource(resource: Path):
            if resource.is_symlink():
                resource.unlink()
            elif resource.is_dir():
                shutil.rmtree(resource)
            else:
                os.remove(resource)

        resources_folder = self.pipeline.path / "resources"
        if resources_folder.exists():
            resources.insert(0, Path("resources"))
        if self.verbose and resources:
            typer.secho(
                f"Copying {len(resources)} resources to working directory...",
                fg=typer.colors.MAGENTA,
            )
        try:
            for resource in resources:
                abs_path = self.pipeline.path / resource
                destination = Path(".") / resource.name
                if not destination.exists():
                    # make sure you don't delete files that are already there...
                    copy_resource(abs_path, destination, symlink=symlink_resources)
                    copied_resources.append(destination)
                elif self.verbose:
                    typer.secho(
                        f"  - Resource '{resource.name}' already exists! Skipping...",
                        fg=typer.colors.MAGENTA,
                    )
            yield
        finally:
            if not cleanup:
                return
            for copied_resource in copied_resources:
                if copied_resource.exists():
                    if self.verbose:
                        typer.secho(
                            f"Deleting '{copied_resource.name}' resource...",
                            fg=typer.colors.MAGENTA,
                        )
                    remove_resource(copied_resource)

def check_command_available(command: str):
    """
    Check if a command is available.
    Args:
      command (str): The command to check.
    Returns:
      bool: True if the command is available, False otherwise.
    Examples:
      >>> CLI.check_command_available('ls')
    """
    import shutil

    return shutil.which(command) is not None

def parse_config_monkeypatch(args):
    """Monkeypatch the parse_config function from snakemake."""
    import yaml
    import snakemake
    import re

    def _yaml_safe_load(s):
        """Load yaml string safely."""
        s = s.replace(": None", ": null")
        return yaml.load(s, Loader=yaml.SafeLoader)

    parsers = [int, float, snakemake._bool_parser, _yaml_safe_load, str]
    config = dict()
    if args.config is not None:
        valid = re.compile(r"[a-zA-Z_]\w*$")
        for entry in args.config:
            key, val = snakemake.parse_key_value_arg(
                entry,
                errmsg="Invalid config definition: Config entries have to be defined as name=value pairs.",
            )
            if not valid.match(key):
                raise ValueError(
                    "Invalid config definition: Config entry must start with a valid identifier."
                )
            v = None
            if val == "" or val == "None":
                snakemake.update_config(config, {key: v})
                continue
            for parser in parsers:
                try:
                    v = parser(val)
                    # avoid accidental interpretation as function
                    if not callable(v):
                        break
                except:
                    pass
            assert v is not None
            snakemake.update_config(config, {key: v})
    return config



from typing import List
from pathlib import Path
from datetime import datetime
import typer
import sys
import collections  # MutableMapping import hack

if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    from collections.abc import MutableMapping
else:
    from collections import MutableMapping
from snk.cli.options import Option


def flatten(d, parent_key="", sep=":"):
    """
    Flattens a nested dictionary.
    Args:
      d (dict): The dictionary to flatten.
      parent_key (str, optional): The parent key of the dictionary. Defaults to ''.
      sep (str, optional): The separator to use between keys. Defaults to ':'.
    Returns:
      dict: A flattened dictionary.
    Examples:
      >>> d = {'a': {'b': 1, 'c': 2}, 'd': 3}
      >>> flatten(d)
      {'a:b': 1, 'a:c': 2, 'd': 3}
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def convert_key_to_snakemake_format(key, value, sep=":"):
    """
    Convert key to a format that can be passed over the cli to snakemake
    """
    result_dict = {}
    parts = key.split(sep)
    current_dict = result_dict

    for part in parts[:-1]:
        current_dict = current_dict.setdefault(part, {})

    current_dict[parts[-1]] = value

    return result_dict


def serialise(d):
    """
    Serialises a data structure into a string.
    Args:
      d (any): The data structure to serialise.
    Returns:
      any: The serialised data structure.
    Examples:
      >>> serialise({'a': 1, 'b': 2})
      {'a': '1', 'b': '2'}
    """
    if isinstance(d, Path) or isinstance(d, datetime):
        return str(d)

    if isinstance(d, list):
        return [serialise(x) for x in d]

    if isinstance(d, dict):
        for k, v in d.items():
            d.update({k: serialise(v)})

    # return anything else, like a string or number
    return d


def parse_config_args(args: List[str], options: List[Option]):
    """
    Parses a list of arguments and a list of options.
    Args:
      args (List[str]): A list of arguments.
      options (List[Option]): A list of options.
    Returns:
      (List[str], List[dict]): A tuple of parsed arguments and config.
    Examples:
      >>> parse_config_args(['-name', 'John', '-age', '20'], [{'name': 'name', 'default': '', 'help': '', 'type': 'str', 'required': True}, {'name': 'age', 'default': '', 'help': '', 'type': 'int', 'required': True}])
      (['John', '20'], [{'name': 'name', 'John'}, {'age': 20}])
    """
    names = [op.name for op in options]
    config = []
    parsed: List[str] = []
    flag = None
    for arg in args:
        if flag:
            name = flag.lstrip("-")
            op = next(op for op in options if op.name == name)
            if op.updated is False and op.default == serialise(arg):
                # skip args that don't change
                flag = None
                continue
            if ":" in op.original_key:
                samkemake_format_config = convert_key_to_snakemake_format(
                    op.original_key, arg
                )
                name = list(samkemake_format_config.keys())[0]
                arg = samkemake_format_config[name]
            config.append({name: serialise(arg)})
            flag = None
            continue
        if arg.startswith("-") and arg.lstrip("-") in names:
            flag = arg
            continue
        parsed.append(arg)
    parsed.sort()
    return parsed, config


def get_default_type(v):
    default_type = type(v)
    if default_type == list and len(v) > 0:
        return f"List[{type(v[0]).__name__}]"
    return str(default_type.__name__)


def dag_filetype_callback(ctx: typer.Context, file: Path):
    allowed = [".pdf", ".png", ".svg"]
    if ctx.resilient_parsing or not file:
        return
    if file.suffix not in allowed:
        raise typer.BadParameter(f"Dag file suffix must be one of {','.join(allowed)}!")
    return file



import pytest
from snk.cli.config.config import SnkConfig


def test_snk_config_creation():
    snk_config = SnkConfig()
    assert snk_config.tagline == "A Snakemake pipeline CLI generated with Snk"
    assert snk_config.font == "small"
    assert snk_config.resources == []
    assert snk_config.cli == {}
    assert snk_config.symlink_resources == False
    assert snk_config.version == None


def test_validate_resources_with_existing_files(tmp_path):
    resources = [tmp_path / "file1", tmp_path / "file2"]
    for resource in resources:
        resource.touch()
    snk_config = SnkConfig()
    # Should not raise an exception
    snk_config.validate_resources(resources)


def test_validate_resources_with_missing_files(tmp_path):
    resources = [tmp_path / "file1", tmp_path / "missing_file"]
    resources[0].touch()
    snk_config = SnkConfig()
    with pytest.raises(FileNotFoundError):
        snk_config.validate_resources(resources)


def test_add_resources(tmp_path):
    resources = [tmp_path / "file1", tmp_path / "file2"]
    for resource in resources:
        resource.touch()
    snk_config = SnkConfig()
    snk_config.add_resources(resources)
    assert len(snk_config.resources) == 2


def test_from_path_with_existing_file(tmp_path):
    config_file = tmp_path / "snk.yaml"
    config_file.touch()
    config_file.write_text("logo: test_logo")
    snk_config = SnkConfig.from_path(config_file)
    assert snk_config._snk_config_path == config_file

def test_missing_file(tmp_path):
    config_file = tmp_path / "snk.yaml"
    with pytest.raises(FileNotFoundError):
        SnkConfig.from_path(config_file)

def test_empty_file(tmp_path):
    config_file = tmp_path / "snk.yaml"
    config_file.touch()
    with pytest.raises(ValueError):
        SnkConfig.from_path(config_file)

def test_invalid_yaml(tmp_path):
    config_file = tmp_path / "snk.yaml"
    config_file.touch()
    config_file.write_text("logo: test_logo\ninvalid_yaml")
    with pytest.raises(Exception):
        SnkConfig.from_path(config_file)

def test_from_dir_with_existing_file(tmp_path):
    config_file = tmp_path / ".snk"
    config_file.touch()
    config_file.write_text("logo: test_logo")
    snk_config = SnkConfig.from_pipeline_dir(tmp_path)
    assert snk_config._snk_config_path == config_file


def test_from_path_with_missing_file(tmp_path):
    config_file = tmp_path / "missing_file.yaml"
    # assert it raises FileNotFoundError
    with pytest.raises(FileNotFoundError):
        SnkConfig.from_path(config_file)


def test_to_yaml(tmp_path):
    snk_config = SnkConfig(art="test_art", logo="test_logo")
    yaml_file = tmp_path / "snk.yaml"
    snk_config.to_yaml(yaml_file)
    assert yaml_file.exists()


def test_save(tmp_path):
    config_file = tmp_path / "snk.yaml"
    snk_config = SnkConfig(_snk_config_path=config_file)
    snk_config.art = "new_art"
    snk_config.save()
    saved_snk_config = SnkConfig.from_path(config_file)
    assert saved_snk_config.art == "new_art"


def test_version_from_about_file(tmp_path):
    config_file = tmp_path / "snk.yaml"
    about_file = tmp_path / "__about__.py"
    about_file.touch()
    about_file.write_text("__version__ = '0.0.1'")
    config_file.write_text("version: __about__.py")
    snk_config = SnkConfig.from_path(config_file)
    assert snk_config.version == "0.0.1"

def test_version_from_about_file_with_missing_file(tmp_path):
    config_file = tmp_path / "snk.yaml"
    about_file = tmp_path / "__about__.py"
    config_file.write_text("version: __about__.py")
    with pytest.raises(FileNotFoundError):
        SnkConfig.from_path(config_file)


from pathlib import Path
from .cli import CLI


def create_cli(p):
    """
    This is the interface to create the dynamic CLI.
    Changing this function name will break backwards compatibility.
    """
    pipeline_dir_path = Path(p)
    cli = CLI(pipeline_dir_path)
    cli()



import typer
from typing import List, Callable
from inspect import signature, Parameter
from makefun import with_signature

from snk.cli.options import Option
import sys


class DynamicTyper:
    app: typer.Typer

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.app = typer.Typer()

    def __call__(self):
        """
        Invoke the CLI.
        Side Effects:
          Invokes the CLI.
        Examples:
          >>> CLI(Path('/path/to/pipeline'))()
        """
        self.app()

    def register_default_command(self, command: Callable, **command_kwargs) -> None:
        """
        Register a default command to the CLI.
        Args:
          command (Callable): The command to register.
        Side Effects:
          Registers the command to the CLI.
        Examples:
          >>> CLI.register_default_command(my_command)
        """
        from makefun import with_signature
        from inspect import signature, Parameter

        command_signature = signature(command)
        params = list(command_signature.parameters.values())
        has_ctx = any([p.name == "ctx" for p in params])
        if not has_ctx:
            params.insert(
                0,
                Parameter(
                    "ctx",
                    kind=Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=typer.Context,
                ),
            )
            command_signature = command_signature.replace(parameters=params)

        @with_signature(command_signature)
        def wrapper(ctx: typer.Context, *args, **kwargs):
            if ctx.invoked_subcommand is None:
                if has_ctx:
                    return command(ctx, *args, **kwargs)
                return command(*args, **kwargs)

        self.register_callback(wrapper, invoke_without_command=True, **command_kwargs)

    def register_command(
        self, command: Callable, dynamic_options=None, **command_kwargs
    ) -> None:
        """
        Register a command to the CLI.
        Args:
          command (Callable): The command to register.
        Side Effects:
          Registers the command to the CLI.
        Examples:
          >>> CLI.register_command(my_command)
        """
        if dynamic_options is not None:
            command = self.add_dynamic_options(command, dynamic_options)
        if isinstance(command, DynamicTyper):
            self.app.registered_commands.extend(command.app.registered_commands)
        else:
            self.app.command(**command_kwargs)(command)

    def register_callback(self, command: Callable, **command_kwargs) -> None:
        """
        Register a callback to the CLI.
        Args:
          command (Callable): The callback to register.
        Side Effects:
          Registers the callback to the CLI.
        Examples:
          >>> CLI.register_callback(my_callback)
        """
        self.app.callback(**command_kwargs)(command)

    def register_group(self, group: "DynamicTyper", **command_kwargs) -> None:
        """
        Register a subcommand group group to the CLI.
        Args:
          group (DynamicTyper): The subcommand group to register.
        Side Effects:
          Registers the subcommand group to the CLI.
        Examples:
          >>> CLI.register_app(my_group)
        """
        self.app.add_typer(group.app, **command_kwargs)

    def _create_cli_parameter(self, option: Option):
        """
        Creates a parameter for a CLI option.
        Args:
          option (Option): An Option object containing the option's name, type, required status, default value, and help message.
        Returns:
          Parameter: A parameter object for the CLI option.
        Examples:
          >>> option = Option(name='foo', type='int', required=True, default=0, help='A number')
          >>> create_cli_parameter(option)
          Parameter('foo', kind=Parameter.POSITIONAL_OR_KEYWORD, default=typer.Option(..., help='[CONFIG] A number'), annotation=int)
        """
        return Parameter(
            option.name,
            kind=Parameter.POSITIONAL_OR_KEYWORD,
            default=typer.Option(
                ... if option.required else option.default,
                *[option.flag, option.short_flag] if option.short else [],
                help=f"[CONFIG] {option.help}",
            ),
            annotation=option.type,
        )

    def check_if_option_passed_via_command_line(self, option: Option):
        """
        Check if an option is passed via the command line.
        Args:
          option (Option): An Option object containing the option's name, type, required status, default value, and help message.
        Returns:
          bool: Whether the option is passed via the command line.
        """
        if option.flag in sys.argv:
            return True
        elif option.short and option.short_flag in sys.argv:
            return True
        return False

    def add_dynamic_options(self, func: Callable, options: List[Option]):
        """
        Function to add dynamic options to a command.
        Args:
          command (Callable): The command to which the dynamic options should be added.
          options (List[dict]): A list of dictionaries containing the option's name, type, required status, default value, and help message.
        Returns:
          Callable: A function with the dynamic options added.
        Examples:
          >>> my_func = add_dynamic_options_to_function(my_func, [{'name': 'foo', 'type': 'int', 'required': True, 'default': 0, 'help': 'A number'}])
          >>> my_func
          <function my_func at 0x7f8f9f9f9f90>
        """
        func_sig = signature(func)
        params = list(func_sig.parameters.values())
        for op in options[::-1]:
            params.insert(1, self._create_cli_parameter(op))
        new_sig = func_sig.replace(parameters=params)

        @with_signature(func_signature=new_sig, func_name=func.__name__)
        def func_wrapper(*args, **kwargs):
            """
            Wraps a function with dynamic options.
            Args:
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.
            Returns:
                Callable: A wrapped function with the dynamic options added.
            Notes:
                This function is used in the `add_dynamic_options_to_function` function.
            """
            flat_config = None

            if kwargs.get("configfile"):
                import snakemake
                from .utils import flatten

                snakemake_config = snakemake.load_configfile(kwargs["configfile"])
                flat_config = flatten(snakemake_config)

            for option in options:

                def add_option_to_args():
                    kwargs["ctx"].args.extend([f"--{option.name}", kwargs[option.name]])

                passed_via_command_line = self.check_if_option_passed_via_command_line(
                    option
                )

                # If no config file provided, or an option is passed, add it to the arguments
                if flat_config is None or passed_via_command_line:
                    add_option_to_args()
                # If a config file is provided and the option key isn't in it, add the option to the arguments
                elif flat_config and option.original_key not in flat_config:
                    add_option_to_args()

            kwargs = {
                k: v for k, v in kwargs.items() if k in func_sig.parameters.keys()
            }
            return func(*args, **kwargs)

        return func_wrapper

    def error(self, msg, exit=True):
        """
        Logs an error message (red) and exits (optional).
        Args:
          msg (str): The error message to log.
          exit (bool): Whether to exit after logging the error message.
        """
        typer.secho(msg, fg="red")
        if exit:
            raise typer.Exit(1)

    def log(self, msg):
        """
        Logs a message (yellow).
        Args:
          msg (str): The message to log.
        """
        typer.secho(msg, fg="yellow")

    def echo(self, msg):
        """
        Prints a message.
        Args:
          msg (str): The message to print.
        """
        typer.echo(msg)



from pathlib import Path
import sys
from typing import Optional
from git import Repo, InvalidGitRepositoryError

from snk.cli.config.utils import get_version_from_config


class Pipeline:
    """
    Represents a pipeline.
    Attributes:
      path (Path): The path to the pipeline.
      repo (Repo): The git repository of the pipeline.
      name (str): The name of the pipeline.
    """

    def __init__(self, path: Path) -> None:
        """
        Initializes a Pipeline object.
        Args:
            path (Path): The path to the pipeline.
        Returns:
            None
        Notes:
            Initializes the `repo` and `name` attributes.
        """
        self.path = path
        if path.is_symlink():  # editable mode
            self.repo = None
        else:
            try:
                self.repo = Repo(path)
            except InvalidGitRepositoryError:
                self.repo = None
        self.name = self.path.name

    @property
    def tag(self):
        """
        Gets the tag of the pipeline.
        Returns:
            str: The tag of the pipeline, or None if no tag is found.
        """
        try:
            # TODO: default to commit
            tag = self.repo.git.describe(["--tags", "--exact-match"])
        except Exception:
            tag = None
        return tag
    
    @property
    def commit(self):
        """
        Gets the commit SHA of the pipeline.
        Returns:
            str: The commit SHA of the pipeline.
        """
        try:
            sha = self.repo.head.object.hexsha
            commit = self.repo.git.rev_parse(sha, short=8)
        except Exception:
           commit = None
        return commit

    @property
    def version(self):
        """
        Gets the version of the pipeline.
        Returns:
            str: The version of the pipeline, or None if no version is found.
        """
        version = None
        if (self.path / "snk.yaml").exists():
            version = get_version_from_config(self.path / "snk.yaml")
        if version is None and not self.tag:
            version = self.commit
        else:
            version = self.tag
        return version if version else "latest"

    @property
    def executable(self):
        """
        Gets the executable of the pipeline.
        Returns:
            Path: The path to the pipeline executable.
        """
        pipeline_bin_dir = self.path.parent.parent / "bin"
        name = self.name
        if sys.platform.startswith("win"):
            name += ".exe"
        return pipeline_bin_dir / name

    @property
    def editable(self):
        """Is the pipeline editable?"""
        return self.path.is_symlink()

    def _find_folder(self, name) -> Optional[Path]:
        """Search for folder"""
        if (self.path / name).exists():
            return self.path / name
        if (self.path / "workflow" / name).exists():
            return self.path / "workflow" / name
        return None

    @property
    def profiles(self):
        pipeline_profile_dir = self._find_folder("profiles")
        if pipeline_profile_dir:
            return [p for p in pipeline_profile_dir.glob("*") if p.is_dir()]
        return []

    @property
    def environments(self):
        pipeline_environments_dir = self._find_folder("envs")
        if pipeline_environments_dir:
            return [e for e in pipeline_environments_dir.glob("*.yaml")] + [
                e for e in pipeline_environments_dir.glob("*.yml")
            ]
        return []

    @property
    def scripts(self):
        pipeline_environments_dir = self._find_folder("scripts")
        if pipeline_environments_dir:
            return [s for s in pipeline_environments_dir.iterdir() if s.is_file()]
        return []



from snk import Nest
from pathlib import Path


def test_init(bin_dir, snk_home):
    nest = Nest(snk_home, bin_dir=bin_dir)
    assert nest.snk_pipelines_dir == Path(snk_home) / "pipelines"
    for path in [nest.snk_pipelines_dir, nest.bin_dir, nest.snk_home]:
        assert path.exists()


def test_download(nest: Nest):
    path = nest.download(
        "https://github.com/Wytamma/snk-basic-pipeline.git", "rna-seq-star-deseq2"
    )
    expected_location = nest.snk_pipelines_dir / "rna-seq-star-deseq2"
    assert (expected_location).exists()
    assert path == expected_location


def test_create_package(nest: Nest):
    test_pipeline_path = nest.snk_pipelines_dir / "pipeline-name"
    test_pipeline_path.mkdir()
    path = nest.create_executable(
        pipeline_path=test_pipeline_path, name=test_pipeline_path.name
    )
    assert path == nest.snk_home / "bin" / "pipeline-name"


def test_install(nest: Nest):
    pipeline = nest.install("tests/data/pipeline")
    assert pipeline.name == "pipeline"
    assert len(nest.pipelines) == 1


def test_install_custom_name(nest: Nest):
    pipeline = nest.install("tests/data/pipeline", name="custom-name")
    assert pipeline.name == "custom-name"


def test_link_pipeline_executable_to_bin(nest: Nest):
    pipeline_executable_path = Path("tests/data/bin/snk-basic-pipeline")
    executable_path = nest.link_pipeline_executable_to_bin(pipeline_executable_path)
    assert executable_path.exists() is True and executable_path.is_symlink() is True


def test_uninstall(nest: Nest):
    pass



from dataclasses import dataclass
from pathlib import Path
from snakemake import (
    Workflow,
    update_config,
    load_configfile,
    dict_to_key_value_args,
    common,
)
from snakemake.persistence import Persistence
import os


def create_workflow(
    snakefile,
    cache=None,
    lint=None,
    cores=1,
    nodes=None,
    max_threads=None,
    resources=dict(),
    overwrite_threads=None,
    overwrite_scatter=None,
    overwrite_resource_scopes=None,
    default_resources=None,
    overwrite_resources=None,
    config=dict(),
    configfiles=None,
    config_args=None,
    workdir=None,
    printshellcmds=False,
    rerun_triggers=["mtime", "params", "input", "software-env", "code"],
    conda_cleanup_envs=False,
    latency_wait=3,
    print_compilation=False,
    debug=False,
    all_temp=False,
    jobscript=None,
    overwrite_shellcmd=None,
    restart_times=0,
    attempt=1,
    verbose=False,
    use_conda=False,
    use_singularity=False,
    use_env_modules=False,
    singularity_args="",
    conda_frontend="conda",
    conda_prefix=None,
    conda_cleanup_pkgs=None,
    list_conda_envs=False,
    singularity_prefix=None,
    shadow_prefix=None,
    scheduler="ilp",
    scheduler_ilp_solver=None,
    mode=common.Mode.default,
    wrapper_prefix=None,
    default_remote_prefix="",
    assume_shared_fs=True,
    keep_metadata=True,
    edit_notebook=None,
    envvars=None,
    overwrite_groups=None,
    group_components=None,
    max_inventory_wait_time=20,
    execute_subworkflows=True,
    conda_not_block_search_path_envvars=False,
    scheduler_solver_path=None,
    conda_base_path=None,
    local_groupid="local",
):
    overwrite_config = dict()
    if configfiles is None:
        configfiles = []
    for f in configfiles:
        # get values to override. Later configfiles override earlier ones.
        update_config(overwrite_config, load_configfile(f))
    # convert provided paths to absolute paths
    configfiles = list(map(os.path.abspath, configfiles))

    # directly specified elements override any configfiles
    if config:
        update_config(overwrite_config, config)
        if config_args is None:
            config_args = dict_to_key_value_args(config)

    if workdir:
        if not os.path.exists(workdir):
            os.makedirs(workdir)
        workdir = os.path.abspath(workdir)
        os.chdir(workdir)

    workflow = Workflow(
        snakefile=snakefile,
        rerun_triggers=rerun_triggers,
        jobscript=jobscript,
        overwrite_shellcmd=overwrite_shellcmd,
        overwrite_config=overwrite_config,
        overwrite_workdir=workdir,
        overwrite_configfiles=configfiles,
        overwrite_clusterconfig=dict(),
        overwrite_threads=overwrite_threads,
        max_threads=max_threads,
        overwrite_scatter=overwrite_scatter,
        overwrite_groups=overwrite_groups,
        overwrite_resources=overwrite_resources,
        overwrite_resource_scopes=overwrite_resource_scopes,
        group_components=group_components,
        config_args=config_args,
        debug=debug,
        verbose=verbose,
        use_conda=use_conda or list_conda_envs or conda_cleanup_envs,
        use_singularity=use_singularity,
        use_env_modules=use_env_modules,
        conda_frontend=conda_frontend,
        conda_prefix=conda_prefix,
        conda_cleanup_pkgs=conda_cleanup_pkgs,
        singularity_prefix=singularity_prefix,
        shadow_prefix=shadow_prefix,
        singularity_args=singularity_args,
        scheduler_type=scheduler,
        scheduler_ilp_solver=scheduler_ilp_solver,
        mode=mode,
        wrapper_prefix=wrapper_prefix,
        printshellcmds=printshellcmds,
        restart_times=restart_times,
        attempt=attempt,
        default_remote_provider=None,
        default_remote_prefix=default_remote_prefix,
        run_local=True,
        assume_shared_fs=assume_shared_fs,
        default_resources=default_resources,
        cache=cache,
        cores=cores,
        nodes=nodes,
        resources=resources,
        edit_notebook=edit_notebook,
        envvars=envvars,
        max_inventory_wait_time=max_inventory_wait_time,
        conda_not_block_search_path_envvars=conda_not_block_search_path_envvars,
        execute_subworkflows=execute_subworkflows,
        scheduler_solver_path=scheduler_solver_path,
        conda_base_path=conda_base_path,
        check_envvars=not lint,  # for linting, we do not need to check whether requested envvars exist
        all_temp=all_temp,
        local_groupid=local_groupid,
        keep_metadata=keep_metadata,
        latency_wait=latency_wait,
    )

    @dataclass
    class PersistenceMock(Persistence):
        """
        Mock for workflow.persistence
        """

        conda_env_path: Path = None
        _metadata_path: Path = None
        _incomplete_path: Path = None
        shadow_path: Path = None
        conda_env_archive_path: Path = None
        container_img_path: Path = None
        aux_path: Path = None

    workflow.persistence = PersistenceMock(
        conda_env_path=Path(conda_prefix).resolve() if conda_prefix else None,
        conda_env_archive_path=os.path.join(Path(".snakemake"), "conda-archive"),
    )
    return workflow



from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class Option:
    name: str
    original_key: str
    default: Any
    updated: bool
    help: str
    type: Any
    required: bool
    short: Optional[str]
    flag: str
    short_flag: Optional[str]



from types import SimpleNamespace
import typer
from pathlib import Path
from typing import Optional, List
from rich import print
from .nest import Nest
from .errors import PipelineExistsError, PipelineNotFoundError
from .__about__ import __version__


app = typer.Typer()

SNK_HOME = None
SNK_BIN = None

def _print_snk_version(self, ctx: typer.Context, value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()

# fmt: off
@app.callback(context_settings={"help_option_names": ["-h", "--help"]})
def callback(
    ctx: typer.Context,
    home: Optional[Path] = typer.Option(
            None, 
            envvar="SNK_HOME", 
            dir_okay=True, 
            file_okay=False, 
            exists=True, 
            help="Overrides default snk location. Pipelines will be installed in $SNK_HOME/pipelines."
        ),
    bin: Optional[Path] = typer.Option(
            None, 
            envvar="SNK_BIN", 
            dir_okay=True, 
            file_okay=False, 
            exists=True, 
            help="Overrides location of pipeline installations. Pipelines are symlinked here."
        ),
    version: Optional[bool] = typer.Option(
        None,
        "-v",
        "--version",
        help="Show the version and exit.",
        is_eager=True,
        callback=_print_snk_version,
        show_default=False,
        ),
    ):
    """\b
        _            _             _        
       / /\         /\ \     _    /\_\      
      / /  \       /  \ \   /\_\ / / /  _   
     / / /\ \__   / /\ \ \_/ / // / /  /\_\ 
    / / /\ \___\ / / /\ \___/ // / /__/ / / 
    \ \ \ \/___// / /  \/____// /\_____/ /  
     \ \ \     / / /    / / // /\_______/   
 _    \ \ \   / / /    / / // / /\ \ \      
/_/\__/ / /  / / /    / / // / /  \ \ \     
\ \/___/ /  / / /    / / // / /    \ \ \    
 \_____\/   \/_/     \/_/ \/_/      \_\_\   
 \b
 \n
 Snakemake pipeline management system
    """
    ctx.obj = SimpleNamespace(snk_home = home, snk_bin = bin)
# fmt: on


@app.command()
def install(
    ctx: typer.Context,
    pipeline: str = typer.Argument(
        ..., help="Path, URL or Github name (user/repo) of the pipeline to install."
    ),
    name: Optional[str] = typer.Option(
        None,
        "--name",
        "-n",
        help="Rename the pipeline (this name will be used to call the CLI.)",
    ),
    tag: Optional[str] = typer.Option(
        None,
        "--tag",
        "-t",
        help="Tag (version) of the pipeline to install. Can specify a branch name, or tag. If None the latest commit will be installed.",
    ),
    commit: Optional[str] = typer.Option(
        None,
        "--commit",
        "-c",
        help="Commit (SHA) of the pipeline to install. If None the latest commit will be installed.",
    ),
    config: Optional[Path] = typer.Option(
        None, help="Specify a non-standard config location."
    ),
    resource: Optional[List[Path]] = typer.Option(
        [],
        help="Specify resources additional to the resources folder required by the pipeline (copied to working dir at runtime).",
    ),
    no_conda: bool = typer.Option(
        False,
        "--no-conda",
        help="Do not use conda environments by default.",
    ),
    force: Optional[bool] = typer.Option(
        False, "--force", "-f", help="Force install (overwrites existing installs)."
    ),
    editable: Optional[bool] = typer.Option(
        False,
        "--editable",
        "-e",
        help="Whether to install the pipeline in editable mode.",
    ),
):
    """
    Install a pipeline.
    """
    nest = Nest(snk_home=ctx.obj.snk_home, bin_dir=ctx.obj.snk_bin)
    if not nest.bin_dir_in_path():
        bin_dir_yellow = typer.style(nest.bin_dir, fg=typer.colors.YELLOW, bold=False)
        typer.echo(f"Please add SNK_BIN to your $PATH: {bin_dir_yellow}")
    if not Path(pipeline).exists() and not pipeline.startswith("http"):
        pipeline = f"https://github.com/{pipeline}.git"
    try:
        installed_pipeline = nest.install(
            pipeline,
            editable=editable,
            name=name,
            tag=tag,
            commit=commit,
            config=config,
            additional_resources=resource,
            force=force,
            conda=not no_conda,
        )
    except PipelineExistsError as e:
        typer.secho(e, fg="red")
        raise typer.Exit()
    except PipelineNotFoundError as e:
        typer.secho(e, fg="red")
        raise typer.Exit()
    typer.secho(f"Successfully installed {installed_pipeline.name} ({installed_pipeline.version})!", fg="green")


@app.command()
def uninstall(
    ctx: typer.Context,
    name: str = typer.Argument(..., help="Name of the pipeline to uninstall."),
    force: Optional[bool] = typer.Option(
        False, "--force", "-f", help="Force uninstall without asking."
    ),
):
    """
    Uninstall a pipeline.
    """
    nest = Nest(snk_home=ctx.obj.snk_home, bin_dir=ctx.obj.snk_bin)
    try:
        uninstalled = nest.uninstall(name, force=force)
    except PipelineNotFoundError as e:
        typer.secho(e, fg="red")
        raise typer.Exit(1)
    if uninstalled:
        typer.secho(f"Successfully uninstalled {name}!", fg="green")


# @app.command()
# def update():
#     """
#     Update a pipeline.
#     """
#     raise NotImplementedError


@app.command()
def list(
    ctx: typer.Context,
):
    """
    List the installed pipelines.
    """
    nest = Nest(snk_home=ctx.obj.snk_home, bin_dir=ctx.obj.snk_bin)
    try:
        pipelines = nest.pipelines
    except FileNotFoundError:
        pipelines = []
    pipeline_dir_yellow = typer.style(
        nest.snk_pipelines_dir, fg=typer.colors.YELLOW, bold=False
    )
    typer.echo(f"Found {len(pipelines)} pipelines in {pipeline_dir_yellow}")
    for pipeline in pipelines:
        if pipeline.editable:
            print(
                f'- {pipeline.name} ([bold green]editable[/bold green]) -> "{pipeline.path}"'
            )
            continue
        print(f"- {pipeline.name} ([bold green]{pipeline.version}[/bold green])")


# @app.command()
# def run(
#         pipeline: str = typer.Argument(
#             ..., help="URL or Github name (user/repo) of the pipeline to install."
#         ),
#     ):
#     """
#     Run the pipeline in a temporary environment.
#     """
#     raise NotImplementedError

# @app.command()
# def annotations(config: Path):
#     """Generate annotations defaults from config file"""
#     raise NotImplementedError

# @app.command()
# def create(name: str):
#     """Create a default project that can be installed with snk"""
#     raise NotImplementedError
